/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <execinfo.h>
#include <pwd.h>
#include <syslog.h>
#include <libgen.h>

#include "gulm_defines.h"
#include "config_gulm.h"
#include "config_ccs.h"
#include "myio.h"
#include "utils_ip.h"
#include "utils_verb_flags.h"
#include "utils_tostr.h"
#include "lock.h"
#include "core.h"
#include "ltpx.h"

/*****************************************************************************/
/* bits of data used by the log_*() and die() functions. */
uint32_t verbosity=0; /* default is set below in the cmdline args parsing. */
char *ProgramName;

gulm_config_t gulm_config;

/* these are calculated at run time.
 * I will need to do something in the future for nodes with multiple ips.
 * */
char myName[256] = "\0";
uint32_t myIP = 0;

int leave_std_open = FALSE;
/*****************************************************************************/

/**
 * forker - 
 * @name: 
 * 
 * Returns: int
 */
int forker(char *name, int (*start)(int argc, char **argv),
      int argc, char **argv)
{
   pid_t pid;
   int ret;

   if((pid=fork()) == 0 ) {
#ifdef PROFILEING_CHILD_KLUDGE
		extern void _start (void), etext (void);
		monstartup ((u_long) &_start, (u_long) &etext);
#endif
      /* child */
      closelog();

      ret = start(argc, argv);

      exit(ret);

   }else if( pid > 0 ) {
      /* parent */
      log_msg(lgm_Forking, "Forked %s [%d].\n", name, pid);
   }else {
      /* error */
      log_err("Failed to fork %s. %d:%s\n", name, errno, strerror(errno));
   }
   return 0;
}

/**
 * my_daemonize - 
 *
 * Make sure you call this BEFORE any calls that may use syslog.
 */
void my_daemonize(void)
{/* fork out */
#ifndef DEBUG
   pid_t pid;
   int fd, i;

   if((pid=fork())<0)
      { die(ExitGulm_ExecError,
            "Failed to fork off of parent: %s\n",strerror(errno)); }
   else if(pid!=0) exit(ExitGulm_Ok);

   setsid();
   chdir("/");
   umask(0);

   if( leave_std_open ) {
      for(i=open_max()-1; i>=3; --i) close(i); /* close nearly everything */
      close(0); /* close stdin */
      fd = open("/dev/null", O_RDWR ); /* stdin */
      /* stdout and stderr should still be open to the terminal for debug
       * output.
       */
   }else{
      for(i=open_max()-1; i>=0; --i) close(i); /* close everything */
      fd = open("/dev/null", O_RDWR ); /*stdin*/
      dup(fd); /*stdout*/
      dup(fd); /*stderr*/
   }
   fprintf(stderr, "stderr was left open.\n");
   printf("stdout was left open.\n");

   openlog(ProgramName, LOG_PID, LOG_DAEMON);
#endif /*DEBUG*/
}

/**
 * become_nobody - root is bad.
 *
 * we don't need to be root, so don't.
 * well, switch to the user that the config says. try atleast.
 * 
 */
void become_nobody(void)
{
   struct passwd *pw=NULL;
   if( getuid() == 0 ) { /* we are root */
      nice(-10); /* while we're here.... */
      /* is run_as root? */
      if( strcmp(gulm_config.run_as, "root") == 0) return;
      if( (pw = getpwnam(gulm_config.run_as)) != NULL ) {
         setuid(pw->pw_uid);
         setgid(pw->pw_gid);
      }else
      {
         die(ExitGulm_InitFailed,
               "There is no user id \"%s\", cannot continue.\n",
               gulm_config.run_as);
      }
   }
}

/**
 * set_myID - Determin who I am.
 * 
 */
void set_myID(void)
{
   /* cute tricks to set the default name and IP. */
   gethostname(myName, 256);
   /* first see if they want us bound to a certain ip */
   if( load_ip_from_ccs(myName, &myIP) != 0 ) {
      /* If not, then lookup my ip from my full name. */
      if( get_ip_for_name(myName, &myIP) != 0 )
         die(ExitGulm_InitFailed, "Failed to find IP for my name (%s)\n",
               myName);
   }
}


/**
 * usage - how to call this
 */
void usage(void)
{
   printf("Usage:\n");
   printf("\n");
   printf("%s [options]\n",ProgramName);
   printf("\nOptions:\n");
   printf("  -C               Test the config and quit.\n");
   printf("  -e               Leave stdout and stderr open.\n");
   printf("  -h               This text.\n");
   printf("  -V               Version information\n");
   printf("  -v <string>      Set the verbose flags.\n");
   exit(ExitGulm_Usage);
}

/**
 * banner_msg - 
 * 
 * Prints out a bunch of things, mostly stuff so users and support can see
 * right away what's going on. Or something like that.
 * 
 * wonder if i should banner each part of gulm?
 */
void banner_msg(void)
{
   log_msg(lgm_Always, "Starting %s %s. (built " __DATE__" " __TIME__
         ")\n%s\n",
         ProgramName, GULM_RELEASE_NAME, SISTINA_COPYRIGHT);
   log_msg(lgm_Always, "You are running in %s mode.\n",
         gulm_config.fog?"Fail-over":"Standard" );
   log_msg(lgm_Always, "I am (%s) with ip (%s)\n", myName, iptostr(myIP));
}

/**
 * parse_cmdline - 
 * @argc: 
 * @argv: 
 * 
 * Returns: int
 */
int parse_cmdline(int argc, char **argv)
{
   int c, conf_test=FALSE;
   extern char *optarg;
   extern int optind;

   ProgramName = strdup(basename(argv[0]));
   if( ProgramName == NULL ) die(ExitGulm_NoMemory, "Out of Memory.\n");
   set_verbosity("Default", &verbosity);

   while((c = getopt(argc, argv, "m:lv:eCVh"))!= -1){
      switch(c) {
         case 'V':
            printf("%s %s (built " __DATE__ " " __TIME__
                    ")\n%s\n",
                    ProgramName, GULM_RELEASE_NAME, SISTINA_COPYRIGHT);
            exit(ExitGulm_Usage);
            break;
         case 'v':
            set_verbosity(optarg, &verbosity);
            break;
         case 'C':
            conf_test = TRUE;
            break;
         case 'e':
            leave_std_open = TRUE;
            break;
         case 'm':
         case 'l':
            fprintf(stderr, "The -%c flag is deprecated.\n\n", c);
         case 'h':
            usage();
         case ':':
         case '?':
            fprintf(stderr, "Please use '-h' for usage.\n");
            exit(ExitGulm_BadOption);
         default:
            fprintf(stderr,"Bad programmer! You forgot to catch the %c flag\n",
                    c);
            exit(1);
            break;
      }
   }

   if( argv[optind] != NULL) {
      die(ExitGulm_BadOption, "Extra command line parameters. Try again.\n");
   }

   /* scan|load|get|whatever conf */
   if( parse_conf(&gulm_config) != 0) {
      die(ExitGulm_ParseFail,"Errors parsing configuration from ccs: %s\n",
            strerror(errno));
      return -1;
   }

   if( conf_test ) {
      dump_conf(&gulm_config);
      exit(ExitGulm_Ok);
   }

   return 0;
}

/**
 * main - 
 * @argc: 
 * @argv: 
 * 
 * 
 * Returns: int
 */
int main(int argc, char **argv)
{
   /* parse cmdline (and config) */
   if( parse_cmdline(argc, argv) != 0 ) return -1;
   /* I would rather see each part get thier own info out of ccs, rather
    * than getting it all here, and passing it to all.
    * Maybe I'll do that someday.
    *
    * well. core shouldn't need ccs.  it should be able to get all it needs
    * from either cmdline or a tiny conf file.  and both of those should be
    * things that can vary from node to node in the same cluster.
    *
    * Other services may stll need some kind of cluster condifguration
    * system.
    */

   /* daemonize ourselves here */
   become_nobody();

   set_myID();

   my_daemonize();
   banner_msg();

   /* fork core */
   forker("core", core_main, argc, argv);

   /* just a moment...*/
   sleep(1);

   /* fork lt */
   forker("locktable", lt_main, argc, argv);

   /* just a moment...*/
   sleep(1);

   /* fork ltproxy */
   forker("ltpx", ltpx_main, argc, argv);

   /* all done. */

   return 0;
}

/* vim: set ai cin et sw=3 ts=3 : */
